@@ -7,11 +7,15 @@
  *	                   Thomas Schwinge <tschwinge@gnu.org>
  *	Copyright (c) 2007 Thomas Schwinge <tschwinge@gnu.org>
  *
- *	Can be freely distributed and used under the terms of the GNU GPL.
+ *	Can be freely distributed and used under the terms of the GNU GPL v2+
+ *
+ *	SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include <sys/io.h>
 
+#include "i386-io-access.h"
+
 static inline int
 intel_setup_io(struct pci_access *a UNUSED)
 {
