@@ -163,6 +163,11 @@ static struct pci_methods *pci_methods[PCI_ACCESS_MAX] = {
 #else
   NULL,
 #endif
+#ifdef PCI_HAVE_PM_RT_THREAD_SMART_DM
+  &pm_rt_thread_smart_dm,
+#else
+  NULL,
+#endif
 };
 
 // If PCI_ACCESS_AUTO is selected, we probe the access methods in this order
@@ -181,6 +186,7 @@ static int probe_sequence[] = {
   PCI_ACCESS_WIN32_KLDBG,
   PCI_ACCESS_WIN32_SYSDBG,
   PCI_ACCESS_AOS_EXPANSION,
+  PCI_ACCESS_RT_THREAD_SMART_DM,
   // Low-level methods poking the hardware directly
   PCI_ACCESS_ECAM,
   PCI_ACCESS_I386_TYPE1,
@@ -400,6 +406,7 @@ retry:
           pci_set_name_list_path(a, path, 1);
         }
     }
+}
 
 #elif defined PCI_OS_AMIGAOS
 
